@@ -119,12 +119,10 @@ soup_transfer_read_cancel (guint tag)
 {
 	SoupReader *r = GINT_TO_POINTER (tag);
 
-	if (r->processing) return;
+	if (r->processing)
+		return;
 
-	if (r->read_tag)
-		g_source_remove (r->read_tag);
-	if (r->err_tag)
-		g_source_remove (r->err_tag);
+	soup_transfer_read_pause (tag);
 
 	g_byte_array_free (r->recv_buf, r->callback_issued ? FALSE : TRUE);
 
@@ -192,12 +190,15 @@ soup_transfer_read_error_cb (GIOChannel* iochannel,
 		goto CANCELLED;
 	}
 
-	IGNORE_CANCEL (r);
-	if (r->error_cb) (*r->error_cb) (body_started, r->user_data);
-	UNIGNORE_CANCEL (r);
+	if (r->error_cb) {
+		soup_transfer_read_pause (GPOINTER_TO_UINT (r));
+		IGNORE_CANCEL (r);
+		(*r->error_cb) (body_started, r->user_data);
+		UNIGNORE_CANCEL (r);
+	}
 
  CANCELLED:
-	soup_transfer_read_cancel (GPOINTER_TO_INT (r));
+	soup_transfer_read_cancel (GPOINTER_TO_UINT (r));
 
 	return FALSE;
 }
@@ -546,31 +547,59 @@ soup_transfer_read (GIOChannel            *chan,
 	reader->recv_buf = g_byte_array_new ();
 	reader->encoding = SOUP_TRANSFER_UNKNOWN;
 
-	reader->read_tag =
-		g_io_add_watch (chan,
-				G_IO_IN,
-				(GIOFunc) soup_transfer_read_cb,
-				reader);
+	soup_transfer_read_unpause (GPOINTER_TO_UINT (reader));
 
-	reader->err_tag =
-		g_io_add_watch (chan,
-				G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-				(GIOFunc) soup_transfer_read_error_cb,
-				reader);
-
-	return GPOINTER_TO_INT (reader);
+	return GPOINTER_TO_UINT (reader);
 }
 
+void
+soup_transfer_read_pause (guint tag)
+{
+	SoupReader *r = GINT_TO_POINTER (tag);
+
+	g_return_if_fail (tag != 0);
+
+	if (r->read_tag) {
+		g_source_remove (r->read_tag);
+		r->read_tag = 0;
+	}
+	if (r->err_tag) {
+		g_source_remove (r->err_tag);
+		r->err_tag = 0;
+	}
+}
+
+void
+soup_transfer_read_unpause (guint tag)
+{
+	SoupReader *r = GINT_TO_POINTER (tag);
+
+	g_return_if_fail (tag != 0);
+
+	if (!r->read_tag) {
+		r->read_tag = g_io_add_watch (r->channel,
+					      G_IO_IN,
+					      (GIOFunc) soup_transfer_read_cb,
+					      r);
+	}
+	if (!r->err_tag) {
+		r->err_tag = g_io_add_watch (r->channel,
+					     G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+					     (GIOFunc) soup_transfer_read_error_cb,
+					     r);
+	}
+}
+
+
 void
 soup_transfer_write_cancel (guint tag)
 {
 	SoupWriter *w = GINT_TO_POINTER (tag);
 
-	if (w->processing) return;
+	if (w->processing)
+		return;
 
-	if (w->write_tag)
-		g_source_remove (w->write_tag);
-	g_source_remove (w->err_tag);
+	soup_transfer_write_pause (tag);
 
 	g_byte_array_free (w->write_buf, TRUE);
 
@@ -583,12 +612,13 @@ soup_transfer_write_error_cb (GIOChannel* iochannel,
 			      SoupWriter *w)
 {
 	if (w->error_cb) {
+		soup_transfer_write_pause (GPOINTER_TO_UINT (w));
 		IGNORE_CANCEL (w);
 		(*w->error_cb) (w->headers_done, w->user_data);
 		UNIGNORE_CANCEL (w);
 	}
 
-	soup_transfer_write_cancel (GPOINTER_TO_INT (w));
+	soup_transfer_write_cancel (GPOINTER_TO_UINT (w));
 
 	return FALSE;
 }
@@ -769,17 +799,7 @@ create_writer (GIOChannel             *chan,
 	writer->error_cb      = error_cb;
 	writer->user_data     = user_data;
 
-	writer->write_tag =
-		g_io_add_watch (chan,
-				G_IO_OUT,
-				(GIOFunc) soup_transfer_write_cb,
-				writer);
-
-	writer->err_tag =
-		g_io_add_watch (chan,
-				G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-				(GIOFunc) soup_transfer_write_error_cb,
-				writer);
+	soup_transfer_write_unpause (GPOINTER_TO_UINT (writer));
 
 	return writer;
 }
@@ -849,6 +869,10 @@ soup_transfer_write_pause (guint tag)
 		g_source_remove (w->write_tag);
 		w->write_tag = 0;
 	}
+	if (w->err_tag) {
+		g_source_remove (w->err_tag);
+		w->err_tag = 0;
+	}
 }
 
 void  
@@ -865,4 +889,11 @@ soup_transfer_write_unpause (guint tag)
 					(GIOFunc) soup_transfer_write_cb,
 					w);
 	}
+	if (!w->err_tag) {
+		w->err_tag =
+			g_io_add_watch (w->channel,
+					G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+					(GIOFunc) soup_transfer_write_error_cb,
+					w);
+	}
 }
